/* **************************************************************************
 * This file is part of Timothy_benchmark
 *
 * Copyright (c) 2014/15 Maciej Cytowski
 * Copyright (c) 2014/15 ICM, University of Warsaw, Poland
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "global.h"
#include "walk.h"
#include "octree.h"

int init_cells(cell_data **cells,int64_t number_of_cells) {
 int i;
 double x1, x2, x3;
 double z1, z2, z3;
 double r1, r2;
 double l;
 double D=1.0;

 /* allocate cells table */
#ifdef __MIC__
 if (!(cells = (cell_data *) _mm_malloc(number_of_cells*sizeof(cell_data),64))) {
  printf("Not enough memory for cells allocation\n");
  exit(1);
 }
#else
 if (!(*cells = (cell_data *) malloc(number_of_cells*sizeof(cell_data)))) {
  printf("Not enough memory for cells allocation\n");
  exit(1);
 }
#endif

 csize = (1024 / 2) / pow(8.0 * number_of_cells * 2.0, 1.0 / 3.0);
 D = csize * pow(8.0 * number_of_cells, 1.0 / 3.0);
 h = 3.0 * csize;
 h2=h*h;
 h3=h2*h;
 
 /* set cell's initial position and state */
 for (i = 0; i < number_of_cells; i++) {
  r2 = 1.1;
  while (r2 >= 1.0) {
   r1 = 1.1;
   while (r1 == 0 || r1 >= 1.0) {
    x1 = ((double)rand()/RAND_MAX)*2.0-1.0;
    x2 = ((double)rand()/RAND_MAX)*2.0-1.0;
    x3 = ((double)rand()/RAND_MAX)*2.0-1.0;
    r1 = x1 * x1 + x2 * x2 + x3 * x3;
   }
   l = sqrt(-2 * log(r1) / r1);
   z1 = x1 * l;
   z2 = x2 * l;
   z3 = x3 * l;

   r2 = z1 * z1 + z2 * z2 + z3 * z3;
  }

  (*cells)[i].x = z1 * D + 1024.0 / 2.0;
  (*cells)[i].y = z2 * D + 1024.0 / 2.0;
  (*cells)[i].z = z3 * D + 1024.0 / 2.0;
 
  (*cells)[i].size = pow(2.0, -(1.0 / 3.0)) * csize;
  (*cells)[i].gid = (long long)i;
  (*cells)[i].v = 0.0;
  (*cells)[i].density = 0.0;
  (*cells)[i].h = h;
  (*cells)[i].young = 2100.0 + ((double)rand()/RAND_MAX) * 100.0;
  (*cells)[i].halo = 0;
  (*cells)[i].phase = 0;
  (*cells)[i].g1 = 11.0 * (1 + (((double)rand()/RAND_MAX) * 2 - 1) * 0.5);
  (*cells)[i].g2 = 4.0 * (1 + (((double)rand()/RAND_MAX) * 2 - 1) * 0.5);
  (*cells)[i].s = 8.0 * (1 + (((double)rand()/RAND_MAX) * 2 - 1) * 0.5);
  (*cells)[i].m = 1.0 * (1 + (((double)rand()/RAND_MAX) * 2 - 1) * 0.5);
  (*cells)[i].phasetime = 0.0;
  (*cells)[i].tumor = 0;
  (*cells)[i].age = 0;
  (*cells)[i].death = 0;
  (*cells)[i].ctype = 0;
  (*cells)[i].scstage = 0;
 } /* for i */
}

void printBasicInfo()
{
  printf("\nTimothy - Tissue Modelling Framework (benchmark)\n");
  printf(" M.Cytowski, Z.Szymanska\n");
  printf(" ICM, University of Warsaw\n");
  printf(" http://timothy.icm.edu.pl\n\n");
  fflush(stdout);
}

int main(int argc,char* argv[])
{
 int i;
 double t1,t2;
 int number_of_iterations;
 int64_t number_of_cells;
 MIC_ATTR cell_data *cells;       
 if (argc > 1) 
  number_of_cells = atoi(argv[1]);
 else  
  number_of_cells = 1<<26;

 number_of_iterations = 128;

 printBasicInfo();

 printf("Number of cells: %ld\n\n",number_of_cells);

 init_cells(&cells,number_of_cells);

 printf("Iteration\tTree time\tWalk time\n");

 for(i=0;i<number_of_iterations;i++) {
   t1=octBuild(cells,number_of_cells);
   t2=walk(cells,number_of_cells);
   printf("%9d\t%9.4f\t%9.4f\n",i,t1,t2);
 }
 return 0;
}
