/* **************************************************************************
 * This file is part of Timothy_benchmark
 *
 * Copyright (c) 2014/15 Maciej Cytowski
 * Copyright (c) 2014/15 ICM, University of Warsaw, Poland
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>
#include "global.h"
#include "octree.h"

MIC_ATTR static inline int octNodeIntersection(int idx,uintVector3d minLocCode,uintVector3d maxLocCode)
{
  if( maxLocCode.x<=octree[idx].xcode || minLocCode.x>=octree[idx].xlimit ) return 0;
  if( maxLocCode.y<=octree[idx].ycode || minLocCode.y>=octree[idx].ylimit ) return 0;
  if( maxLocCode.z<=octree[idx].zcode || minLocCode.z>=octree[idx].zlimit ) return 0;
  return 1;
}

/*!
 * This function returns the value of the SPH kernel function for given two cells p1 and p2.
 * Kernel function as in:
 * - J.J.Monaghan,"Smoothed Particle Hydrodynamics",Annu.Rev.Astron.Astrophys.,1992,30:543-74
 * - V.Springel,"Smoothed Particle Hydrodynamics in Astrophysics",arXiv:1109.2219v1
 */
MIC_ATTR static inline float sph_kernel(double r)
{
  double u,c=1.0;

  if(r<0.0) return 0.0;

  u=r/h;
  c=8/(M_PI*h3);

  if(u>=0.0 && u<=0.5) return c*(1-6*u*u+6*u*u*u);
  if(u>0.5 && u<=1.0) return c*(2*(1-u)*(1-u)*(1-u));
  if(u>1.0) return 0.0;
  return 0.0;
}

MIC_ATTR double computePotential(cell_data *cells,int p1, int p2)
{
  double pot;
  double dist=0.0;
  double size;
  double x, y, z;
  double xc;
  double D;
  double poisson = 0.33;
  double young;
  int ctype=0;


  if (p1 == p2)
    return 0.0;

  x = cells[p1].x;
  y = cells[p1].y;
  z = cells[p1].z;
  size = cells[p1].size;
  young = cells[p1].young;
  ctype = cells[p1].ctype;

  /* compute the distance between two cells */
  dist =
    sqrt((x - cells[p2].x) * (x - cells[p2].x) +
         (y - cells[p2].y) * (y - cells[p2].y) +
         (z - cells[p2].z) * (z - cells[p2].z));

  if (dist <= h) {

    double r01, r02;
    double area;
    double sc=1.0;

    sc = h3;

    cells[p1].density +=
      sc * (cells[p2].size / csize) * sph_kernel(dist);

    xc = size + cells[p2].size - dist;

    if (xc <= 0.0)
      return 0.0;

    D = 0.75 * ((1.0 - poisson * poisson) / young +
                (1.0 - poisson * poisson / cells[p2].young));

    /* adhesion */
    r01 =
      (size * size - cells[p2].size * cells[p2].size +
       dist * dist) / (2 * dist);
    r02 = dist - r01;

    area =
      M_PI *
      ((size * size * (size - r01) -
        (size * size * size - r01 * r01 * r01) / 3) +
       (cells[p2].size * cells[p2].size * (cells[p2].size - r02) -
        (cells[p2].size * cells[p2].size * cells[p2].size -
         r02 * r02 * r02) / 3));

    /* compute potential */
    pot =
      (2.0 * pow(xc, 5 / 2) / (5.0 * D)) * sqrt((size * cells[p2].size) /
          (size +
           cells[p2].size)) +
      area * 0.1;

    if(ctype==1 && cells[p2].ctype==1) pot = 0.0;

    return pot;

  } else
    return 0.0;

}


MIC_ATTR double walk(cell_data *cells,int64_t number_of_cells)
{
  double t0;

  t0 = omp_get_wtime();  

  if(number_of_cells<=1) return;
  #pragma omp parallel
  {
    int p;

    #pragma omp for schedule(dynamic,64)
    for (p = 0; p < number_of_cells; p++) {
      int64_t cellIdx;
      int newIdx;
      int s;
      int octIndex;
      uintVector3d minLocCode,maxLocCode;
      octHeap octh;
      octHeapInit(&octh);

      cells[p].density = 0.0;
      cells[p].v = 0.0;

      octComputeBox(p,cells,&minLocCode,&maxLocCode);
      octIndex=octLocateRegion(minLocCode,maxLocCode);

      for(s=0; s<8; s++) {
        int idx;
        idx=octree[octIndex].child[s];
        if(idx!=-1 && octNodeIntersection(idx,minLocCode,maxLocCode))
          octHeapPush(&octh,idx);
      }
      while(octh.count>0) {
        int idx;
        idx=octHeapPop(&octh);
        cellIdx=octree[idx].data;
        if(cellIdx>=0) {
          cells[p].v += computePotential(cells,p,cellIdx);
        } else {
          for(s=0; s<8; s++) {
            newIdx=octree[idx].child[s];
            if(newIdx!=-1 && octNodeIntersection(newIdx,minLocCode,maxLocCode))
              octHeapPush(&octh,newIdx);
          }
        }
      } 
      octHeapFree(&octh);
    }
  }
  return omp_get_wtime() - t0;
}

